import os
from pathlib import Path
from t8s.log_config import LogConfig
from behave import given, when, then, model, step
from behave.model import Feature, Scenario
from logging import INFO, DEBUG, WARNING, ERROR, CRITICAL, log

LogConfig().initialize_logger(DEBUG)
logger = LogConfig().getLogger()
print(f'\n\n• The script "environment.py" was loaded ...\n')

def clean_data_dir(CSV_PATH, PARQUET_PATH):
    logger.info(f'clean_data_dir() called ...')
    for diretory in [CSV_PATH, PARQUET_PATH]:
        logger.info(f'clean_data_dir() -> diretory = {diretory}')
        for root, dirs, files in os.walk(diretory):
            for file in files:
                os.remove(os.path.join(root, file))
            for dir in dirs:
                os.rmdir(os.path.join(root, dir))
        for root, dirs, files in os.walk(diretory):
            for file in files:
                print(os.path.join(root, file))
            for dir in dirs:
                print(os.path.join(root, dir))
    return 'data directory empty'

def before_feature(context, feature: Feature):
    # Execute um método uma única vez para todos os cenários de teste da Feature
    logger.info(f'-------------------------------------------------')
    T8S_WORKSPACE_DIR = os.environ.get('T8S_WORKSPACE_DIR', '/Volumes/dev/t8s')
    CSV_PATH_STR:str = os.path.join(T8S_WORKSPACE_DIR, 'data', 'csv')
    CSV_PATH: Path = Path(CSV_PATH_STR)
    PARQUET_PATH_STR:str = os.path.join(T8S_WORKSPACE_DIR, 'data', 'parquet')
    PARQUET_PATH: Path = Path(PARQUET_PATH_STR)
    logger.info(f'before_feature: T8S_WORKSPACE_DIR = {T8S_WORKSPACE_DIR}')
    logger.info(f'before_feature:  CSV_PATH         = {CSV_PATH}')
    logger.info(f'before_feature:  PARQUET_PATH     = {PARQUET_PATH}')
    logger.info(f'-------------------------------------------------')
    # A forma de passar estes dados para os steps seguintes é usando o objeto context
    context.T8S_WORKSPACE_DIR = T8S_WORKSPACE_DIR
    context.CSV_PATH = CSV_PATH
    context.PARQUET_PATH = PARQUET_PATH
    # Atualiza o status no contexto
    context.status = clean_data_dir(context.CSV_PATH, context.PARQUET_PATH)
    logger.info(f'-------------------------------------------------')

def before_scenario(context, scenario: Scenario):
    logger.info(f'-------------------------------------------------')
    logger.info(f'before scenario: {scenario.name}')
    logger.info(f'context.status: {context.status}')
    logger.info(f'-------------------------------------------------')



# -*- coding: utf-8 -*-
# -- FILE: features/steps/example_steps.py

import os
from t8s.log_config import LogConfig
from behave import given, when, then, model, step
from logging import INFO, DEBUG, WARNING, ERROR, CRITICAL

LogConfig().initialize_logger(DEBUG)
logger = LogConfig().getLogger()

@given('we have behave installed')
def check_behave_istalled(context):
    pass

@when('we implement {number} tests')
def check_number(context, number):  # -- NOTE: number is converted into integer
    assert int(number) > 1 or int(number) == 0
    context.tests_count = int(number)
    logger.info(f'STEP: WHEN we implement {number} tests')

@then('behave will test them for us!')
def test(context):
    assert context.failed is False
    assert context.tests_count >= 0

# --------------------------------------------------------------------------------

@given(u'que temos o behave instalado')
def check_behave_istalled_again(context):
    logger.info('STEP: Given que temos o behave instalado')
    logger.info('context' + str(context) + str(type(context)))


@when(u'implementamos {qty} testes')
def check_qty(context, qty):
    logger.info(f'STEP: When implementamos {qty} testes')
    logger.info('context' + str(context) + str(type(context)))

@then(u'o behave vai testar pra gente!')
def teste(context):
    logger.info('STEP: Then o behave vai testar pra gente!')
    logger.info('context' + str(context) + str(type(context)))

# --------------------------------------------------------------------------------

@given(u'a workspace {workspace}')
def set_workspace(context, workspace):
    workspace_dir = os.environ.get('WORKSPACE_DIR', '/Volumes/dev/t8s')
    logger.info(f'STEP: Given a workspace ${workspace} ({workspace_dir})')
    context_workspace = os.environ.get(workspace)
    logger.info(f'STEP: Given a workspace -> context.workspace {context_workspace}')
    context.workspace = context_workspace

@given(u'a valid user logged in')
def select_user(context):
    user = 'fulano'
    logger.info(f'STEP: Given a valid user logged in ({user})')
    context.user = 'fulano'

@given(u'a number {left}')
def a_number(context, left):
    logger.info(f'STEP: Given a number {left}, context.workspace = {context.workspace}')
    context.left = left

@when(u'add a number {right}')
def another_number(context, right):
    logger.info(f'STEP: When add a number {right} value')
    context.right = right

@then(u'the sum is {result}')
def sum(context, result):
    logger.info(f'STEP: Then the sum is {result}')
    # for idx in range(len(context.left)):
    assert int(result) == (int(context.left) + int(context.right))

@given(u'a simple silly step')
def a_silly_step(context):
    logger.info(u'STEP: Given a simple silly step')

@then(u'the last step has a final table')
def final_table(context):
    assert isinstance(context.table, model.Table)
    table: model.Table = context.table
    headings: list[str] = table.headings
    rows: list[list[str]] = [ list(row) for row in table.rows ]
    logger.info(f'STEP: Then the last step has a final table: {table}, {type(table)}')
    logger.info(f'STEP: Then the last step has a final table -> headings: {headings}, {type(headings)}')
    logger.info(f'STEP: Then the last step has a final table -> rows:{rows}, {type(rows)}')
    for idx, row in enumerate(table.rows):
        actual: model.Row = row
        v1 = actual.get(headings[0])
        v2 = actual.get(headings[1])
        v3 = actual.get(headings[2])
        logger.info(f'STEP: Then the last step has a final table -> ' +
            f'table.rows[{idx}]: [ {v1}, {v2}, {v3} ], {type(v1)}, {type(v2)}, {type(v3)}')



from pathlib import Path
from datetime import datetime
from t8s.ts import TimeSerie
from t8s.ts_writer import TSWriter, WriteParquetFile

class Util:
    @staticmethod
    def to_parquet(ts: TimeSerie, path_ts: Path):
        # def write_ts_to_parquet_file(ts, parquet_path, filename: str):
        print(f'Grava a série temporal (formato {ts.format}) em um arquivo parquet {path_ts}')
        context = TSWriter(WriteParquetFile())
        print("Client: Strategy was seted to write Parquet file.")
        context.write(Path(path_ts), ts)
        print(f'\nArquivo {str(path_ts)} gerado à partir da TimeSerie fornecida')



# SPDX-FileCopyrightText: 2023-present João Antonio Ferreira <joao.parana@gmail.com>
#
# SPDX-License-Identifier: MIT
from .__about__ import __version__

# from abc import ABC, abstractmethod
#
# class ITimeSerie(ABC):
#    pass

import re
from enum import Enum
from datetime import datetime
import numpy as np
import pandas as pd

def get_numeric_regex():
    # Definindo a expressão regular para identificar números de ponto flutuante
    pattern = r'[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?'
    return re.compile(pattern)

def get_sample_df(number_of_records: int, first_ts: datetime, time_interval: int) -> tuple[pd.DataFrame, datetime]:
    # Atualmente ignora-se o parâmetro number_of_records.abs
    # TODO: Implementar o uso do parâmetro number_of_records que atualmente é sempre 4
    # Atualmente ignora-se o parâmetro time_interval
    # TODO: Implementar o uso do parâmetro time_interval que atualmente é sempre 1 (hora)
    print('first_ts', first_ts)
    data = {
        'timestamp': [
            first_ts,
            first_ts + pd.Timedelta(hours=1),
            first_ts + pd.Timedelta(hours=2),
            first_ts + pd.Timedelta(hours=3)
        ],
        'temperatura': np.array([25.0, 26.0, 27.0, 23.2], dtype=np.float32),
        'velocidade': [3000, 1100, 1200, 4000],
    }
    # Convertendo os tipos de dado para temperatura e velocidade para
    # np.float32 e np.float32 respectivamente, pois o padrão é np.float64 e np.int64

    temperature_typed_df: np.ndarray = np.array(data['temperatura'], dtype=np.float32)
    data['temperatura'] = temperature_typed_df
    velocity_typed_df: np.ndarray = np.array(data['velocidade'], dtype=np.float32)
    data['velocidade'] = velocity_typed_df
    df = pd.DataFrame(data)
    last_timestamp: datetime = data['timestamp'][-1] # type: ignore
    return (df, last_timestamp)



# -*- coding: utf-8 -*-

from t8s.log_config import LogConfig
from pathlib import Path
from datetime import datetime
import numpy as np
import pandas as pd
import t8s
from t8s import get_sample_df
from t8s.io import IO
from t8s.ts import TimeSerie
from t8s.ts_builder import TSBuilder, ReadParquetFile, ReadCsvFile
from t8s.log_config import LogConfig
from t8s.ts_writer import TSWriter, WriteParquetFile, WriteCsvFile
from logging import INFO, DEBUG, WARNING, ERROR, CRITICAL

if __name__ == "__main__":
    LogConfig().initialize_logger(DEBUG, log_file = 'tests/data_seed.log')
    logger = LogConfig().getLogger()
    logger.info('t8s package version:' + t8s.__version__)
    # Cria uma série temporal multivariada com três atributos: timestamp, temperatura e velocidade
    number_of_records = 4
    time_interval = 1 # hour
    df, last_ts = get_sample_df(number_of_records, datetime(2022, 1, 1, 0, 0, 0), time_interval)
    # Gravando o arquivo CSV
    # df.to_csv(, index=False)
    IO.dataframe_to_csv_file(df, Path('data/csv/ts_01.csv'))
    # Gerando a série temporal
    ts1 = TimeSerie(df, format='wide', features_qty=len(df.columns))
    # Gravando a série temporal em parquet
    path_str: str = 'data/parquet/ts_01.parquet'
    path = Path(path_str)
    print(f'Grava a série temporal (formato {ts1.format}) em um arquivo parquet {path}')
    context = TSWriter(WriteParquetFile())
    print("Client: Strategy was seted to write Parquet file.")
    context.write(Path(path_str), ts1)
    # ---------------------------------------------------------------------------------------------
    # Outro caso de uso
    number_of_records = 4
    time_interval = 1 # hour
    df, last_ts = get_sample_df(number_of_records, last_ts + pd.Timedelta(hours=1), time_interval)
    print('Original:', df)
    # altera a série temporal incluindo NaN e inválidos
    df.iloc[0, 1] = 'Missing'
    df.iloc[1, 1] = np.nan
    df.iloc[2, 2] = np.nan # ATENÇÃO: ao atribuir NaN, o Pandas converte o tipo da coluna de int32 para float64
    df.iloc[3, 1] = 'Bad'
    # df.to_csv('data/csv/ts_02.csv', index=False)
    IO.dataframe_to_csv_file(df, Path('data/csv/ts_02.csv'))

    print('Antes:', df)
    # Gerando a série temporal
    # Filtro valores inválidos, trocando pra NaN
    df.replace(['Configure', 'Missing', 'Bad'], np.nan, inplace=True)
    # Converta os dados numéricos para numpy.float32
    float_cols = df.select_dtypes(include=['float']).columns
    print('float_cols = ', float_cols)
    df[float_cols] = df[float_cols].astype(np.float32)
    print('Depois:', df)
    ts2 = TimeSerie(df, format='wide', features_qty=len(df.columns))
    # Gravando a série temporal em parquet
    path_str: str = 'data/parquet/ts_02.parquet'
    path = Path(path_str)
    print(f'Grava a série temporal (formato {ts2.format}) em um arquivo parquet {path}')
    context = TSWriter(WriteParquetFile())
    print("Client: Strategy was seted to write Parquet file.")
    context.write(Path(path_str), ts2)



"""
Feature: Create a time series set using Dataframe, CSV and Parquet

Value Statement:
  As a data analyst
  I want the hability to create a timeseries set using Dataframe, CSV and Parquet
  So that I can start studying the data right away and propose solutions for the business.

  Background:
    Given that I have a T8S_WORKSPACE_DIR and a bunch of CSV and Parquet files to analise

  Create 2 time series with sample data using Dataframes and save in CSV files at T8S_WORKSPACE/data/csv directory
    Given a start timestamp, a number of records and a time interval
    When I create 2 time series using Dataframes with sample data
    Then I have a time series with the `correct` number of rows and columns, schema and time interval
    And I have a CSV file in T8S_WORKSPACE/data/csv correctelly formated
    # Constraint: The first  Dataframe doesn't have nulls or invalid values, but the second does
"""

import os
from pathlib import Path
from datetime import datetime
import numpy as np
import pandas as pd
from t8s import get_sample_df
from t8s.log_config import LogConfig
from t8s.util import Util
from t8s.io import IO
from t8s.ts import TimeSerie
from t8s.ts_writer import TSWriter, WriteParquetFile
from behave import given, when, then, step
from logging import INFO, DEBUG, WARNING, ERROR, CRITICAL

LogConfig().initialize_logger(DEBUG)
logger = LogConfig().getLogger()

def clean_data_dir():
    logger.info(f'clean_data_dir() called ...')
    pass

@given(u'that I have a T8S_WORKSPACE_DIR and a bunch of CSV and Parquet files to analyze')
def setup(context):
    logger.info(f'-------------------------------------------------')
    logger.info(f'Background @given: T8S_WORKSPACE_DIR = {context.T8S_WORKSPACE_DIR}')
    logger.info(f'Background@given:  CSV_PATH = {context.CSV_PATH}')
    logger.info(f'Background@given:  PARQUET_PATH = {context.PARQUET_PATH}')
    logger.info(f'-------------------------------------------------')
    # A forma de passar estes dados para os steps seguintes é usando o objeto context

@given(u'a start timestamp, a number of records and a time interval')
def create_dataframe(context):
    start_timestamp = datetime(2022, 1, 1, 0, 0, 0)
    number_of_records = 4
    time_interval = 1 # hour
    logger.info(f'@given: -> Building a sample Dataframe')
    dataframe1, last_ts = get_sample_df(number_of_records, start_timestamp, time_interval)
    context.time_interval = time_interval
    context.number_of_records = number_of_records
    context.dataframe1 = dataframe1
    context.last_ts = last_ts
    # altera a série temporal incluindo NaN e inválidos
    dataframe2, last_ts = get_sample_df(number_of_records, last_ts + pd.Timedelta(hours=1), time_interval)
    # print('Original:', dataframe2)
    # altera a série temporal incluindo NaN e inválidos
    dataframe2.iloc[0, 1] = 'Missing'
    dataframe2.iloc[1, 1] = np.nan
    dataframe2.iloc[2, 2] = np.nan # ATENÇÃO: ao atribuir NaN, o Pandas converte o tipo da coluna de int32 para float64
    dataframe2.iloc[3, 1] = 'Bad'
    context.dataframe2 = dataframe2
    context.last_ts = last_ts
    logger.info(f'@given: -> CSV_PATH = {context.CSV_PATH}')

@when(u'I create 2 time series using Dataframes with sample data')
def create_time_series(context):
    logger.info(f'@when: => CSV_PATH = {context.CSV_PATH}')
    # logger.info(f'@when: => context.dataframe1 = \n{context.dataframe1}')
    # logger.info(f'@when: => context.dataframe2 = \n{context.dataframe2}')
    # Gerando a primeira série temporal
    df1 = context.dataframe1
    context.ts1 = TimeSerie(df1, format='wide', features_qty=len(df1.columns))
    df2 = context.dataframe2
    context.ts2 = TimeSerie(df2, format='wide', features_qty=len(df1.columns))

@then(u'I have a time series with the `correct` number of rows and columns, schema and time interval')
def check_schema(context):
    logger.info(f'@then: -> CSV_PATH = {context.CSV_PATH}')
    logger.info(f'@then: -> context.ts1 = \n{context.ts1}')
    logger.info(f'@then: -> context.ts2 = \n{context.ts2}')
    logger.info('------------------------------------------------------------')
    def check_time_interval_equals(ctx, value):
        last_ts = None
        for idx, row in enumerate(ctx.dataframe1.itertuples()):
            if idx == 0:
                last_ts = row[1]
            else:
                interval: pd.Timedelta = row[1] - last_ts
                interval_seconds:int = interval.seconds
                time_interval = interval_seconds / 3600
                assert time_interval == value
                last_ts = row[1]
        logger.info(f'@then: -> time_interval checked')
    check_time_interval_equals(context, context.time_interval)
    assert len(context.ts1.df) == context.number_of_records, 'context.ts1.df must have the correct number of records'
    assert len(context.ts2.df) == context.number_of_records, 'context.ts2.df must have the correct number of records'

@then(u'I have a CSV file in T8S_WORKSPACE/data/csv correctelly formated')
def save_csv(context):
    logger.info(f'@then: => CSV_PATH = {context.CSV_PATH}')
    logger.info(f'@then: => len(context.dataframe1) = \n{len(context.dataframe1)}')
    csv_file_path_str_1 = str(context.CSV_PATH) + '/ts_01.csv'
    IO.dataframe_to_csv_file(context.dataframe1, Path(csv_file_path_str_1))
    csv_file_path_str_2 = str(context.CSV_PATH) + '/ts_02.csv'
    IO.dataframe_to_csv_file(context.dataframe2, Path(csv_file_path_str_2))

@then(u'I have a Parquet file in T8S_WORKSPACE/data/parquet correctelly formated with metadata annotations')
def save_parquet(context):
    logger.info(f'@then: => PARQUET_PATH = {context.PARQUET_PATH}')
    logger.info(f'@then: => len(context.dataframe1) = \n{len(context.dataframe1)}')
    logger.info(f'@then: => len(context.dataframe2) = \n{len(context.dataframe2)}')

    def write_ts_to_parquet_file(ts, parquet_path, filename: str):
        parquet_file_path_str: str = str(parquet_path) + '/' + filename
        path_ts = Path(parquet_file_path_str)
        # Devido a problemas de 'circular import' tivemos que usar a classe Util
        Util.to_parquet(ts, path_ts)
        # print(f'Grava a série temporal (formato {ts.format}) em um arquivo parquet {path_ts}')
        # context = TSWriter(WriteParquetFile())
        # print("Client: Strategy was seted to write Parquet file.")
        # context.write(Path(path_ts), ts)
        # print(f'\nArquivo {str(path_ts)} gerado à partir da TimeSerie fornecida')

    # Grava a série temporal ts1 em parquet
    write_ts_to_parquet_file(context.ts1, context.PARQUET_PATH, 'ts_01.parquet')

    # Grava a série temporal ts2 em parquet
    # write_ts_to_parquet_file(context.ts2, context.PARQUET_PATH, 'ts_02.parquet')



from typing import Any, Callable, Dict, List, Tuple, Union

def given(pattern: Union[str, Callable[..., Any]]) -> Callable[..., Any]: ...
def when(pattern: Union[str, Callable[..., Any]]) -> Callable[..., Any]: ...
def then(pattern: Union[str, Callable[..., Any]]) -> Callable[..., Any]: ...
def step(pattern: Union[str, Callable[..., Any]]) -> Callable[..., Any]: ...



# -*- coding: utf-8 -*-

from __future__ import annotations
from pathlib import Path
from datetime import datetime
import types
import yaml
from abc import ABC, abstractmethod
from typing import Any, Optional, Type, TypeVar

import pandas as pd
import pyarrow as pa         # type: ignore
import pyarrow.parquet as pq # type: ignore

from t8s.log_config import LogConfig

# TODO: Esclarecer duvida coceitual sobre o conceito de feture em séries temporais multivariadas

logger = LogConfig().getLogger()

TS = TypeVar('TS', bound='TimeSerie')

# ts = pd.read_parquet('lixo.parquet', engine='pyarrow')


# Classe abstrata que define a interface para os processamentos de séries temporais
class ITimeSeriesProcessor(ABC):
    def __init__(self):
        self.format: str
        self.features: str
        self.df: pd.DataFrame

    """Converte a série temporal para o formato Long"""

    @abstractmethod
    def to_long(self) -> bool:
        return False

    @abstractmethod
    def to_wide(self) -> bool:
        # Converte a série temporal para o formato Wide
        return False

    @abstractmethod
    def is_univariate(self) -> bool:
        # Verifica se a série temporal é univariada
        return False

    @abstractmethod
    def is_multivariate(self) -> bool:
        # Verifica se a série temporal é multivariada
        return False

    @abstractmethod
    def split(self) -> list[Type['TimeSerie']]:
        # Alternativas para anotar o tipo de retorno:
        # list[TS], list['TimeSerie'] ou list[Optional['TimeSerie']]
        # Cria várias séries temporais univariadas à partir de uma série temporal multivariada
        pass

    # @staticmethod
    # @abstractmethod
    # def join(list_of_ts: list['TimeSerie']) -> TS:
    #     pass


"""Rastreia a proveniência dos ativos de informação, Série Temporal neste caso"""


class IProvenancable(ABC):
    def __init__(self):
        self.format: str
        self.features: str
        self.df: pd.DataFrame

    """Este método adiciona informações de proveniência ao objeto TimeSerie para
    uma transformação específica. Ele recebe como parâmetros o nome da transformação
    e um dicionário com os parâmetros usados na transformação."""

    @abstractmethod
    def add_provenance(self, transformation: str, parameters: dict):
        pass

    @abstractmethod
    def get_provenances(self) -> list[dict[str, Any]]:
        """Este método retorna uma lista de dicionários com as informações de
        proveniência de todas as transformações realizadas no objeto TimeSerie."""
        pass

    @abstractmethod
    def get_provenance_by_transformation(self, transformation: str) -> dict[str, Any]:
        """Este método retorna as informações de proveniência de uma dada transformação
        realizadas no objeto TimeSerie."""
        pass

    @abstractmethod
    def get_last_transformation(self, transformation: str) -> dict[str, Any]:
        """Este método retorna informações da ultima transformação realizada no objeto TimeSerie."""
        pass

    @abstractmethod
    def apply_transformation(self, transformation: str, **kwargs) -> Any:
        """Este método aplica uma transformação ao objeto TimeSerie e retorna um novo
        objeto TimeSerie com as informações de proveniência atualizadas. Ele recebe como
        parâmetros o nome da transformação e os parâmetros específicos da transformação."""
        pass


class ITimeSerie(ITimeSeriesProcessor, IProvenancable):
    """
    Interface que estende as outras interfaces. ë apenas uma Marker Interface
    """
    pass


class TimeSerie(ITimeSerie):
    def __init__(self, *args, format, features_qty, **kwargs):
        assert isinstance(format, str), "format must be a string"
        assert isinstance(features_qty, int), "features_qty must be a int"
        assert format in ['long', 'wide'], "format must be 'long' or 'wide'"
        self.format: str = format
        self.features: str = str(features_qty)
        self.df = pd.DataFrame()
        if len(args) > 0:
            for idx, arg in enumerate(args):
                if idx == 0 and isinstance(arg, pd.DataFrame):
                    self.df = arg
                if idx == 0 and isinstance(arg, dict):
                    if 'content' in arg.keys() and arg['content'] == 'EMPTY':
                        return
                    self.df = pd.DataFrame(*args, **kwargs)
                break # Apenas o primeiro parâmetro precisa de tratamento especial.
        else:
           for key, kwarg_value in kwargs.items():
                if key == 'data' and isinstance(kwarg_value, pd.DataFrame):
                    self.df = kwarg_value
                if key == 'data' and isinstance(kwarg_value, dict):
                    self.df = pd.DataFrame(*args, **kwargs)
                if key == 'data' and isinstance(kwarg_value, str) and kwarg_value == 'EMPTY':
                    return
                break # Apenas o parâmetro 'data' precisa de tratamento especial.

        # TODO: garantir que a primeira coluna seja um Timestamp tanto para o formato long quanto wide
        # Por hora lanço Exception
        if self.df.empty:
            raise Exception('Não foram fornecidos dados para criação da série temporal. Use o método TimeSerie.empty() se desejar criar uma série temporal vazia')
        first_column_type = self.df[self.df.columns[0]].dtype
        if first_column_type != 'datetime64[ns]':
            print(self.df.info())
            raise Exception('A primeira coluna deve ser um Timestamp (datetime)' +
            ' experado ' + 'datetime64[ns]' + ' recebido ' + str(first_column_type))
        logger.debug('Objeto TimeSerie construido com sucesso')

    def __repr__(self):
        columns_types = [type(self.df[col][0]) for col in self.df.columns]
        # Retorna uma representação em string do objeto TimeSerie
        return f'TimeSerie(format={self.format}, features={self.features}, ' + \
               f'df=\n{self.df.__repr__()}) + \ntypes: {columns_types}'

    def to_long(self):
        # Converte a série temporal para o formato Long
        # Implementação aqui
        raise NotImplementedError('Not implemented for long format')

    def to_wide(self):
        # Converte a série temporal para o formato Wide
        # Implementação aqui
        raise NotImplementedError('Not implemented for long format')

    def is_univariate(self):
        # Verifica se a série temporal é univariada
        if self.format == 'long':
            raise NotImplementedError('Not implemented for long format')
        return self.df.columns.size == 2

    def is_multivariate(self):
        # Verifica se a série temporal é multivariada
        if self.format == 'long':
            raise NotImplementedError('Not implemented for long format')
        return self.df.columns.size > 2

    def split(self) -> list[TimeSerie]:  # Alternativa: list['TimeSerie']
        # TODO: garantir que a primeira coluna seja o indice no Dataframe quando o formato for long ou wide
        # TODO: garantir que a primeira coluna seja do tipo Timesamp (datetime) quando o formato for long ou wide
        # Cria várias séries temporais univariadas à partir de uma série temporal multivariada
        result = []
        if self.format == 'long':
            # TODO: Implementar para o caso de formato longo.
            pass
        elif self.format == 'wide':
            # Por contrato a primeira coluna é sempre o timestamp
            timestamp_col = self.df.columns[0]
            for idx, col in enumerate(self.df.columns):
                if idx == 0:
                    continue
                my_df = self.df[[timestamp_col, col]]
                # TODO: criar um novo objeto TimeSerie
                my_ts = TimeSerie(data=my_df, format='wide', features_qty=2)
                logger.debug('---------------------------------------------------')
                logger.debug(f'univariate {idx}' + '\n ' + str(my_df))
                result.append(my_ts)
        else:
            raise Exception('Formato de série temporal não suportado')

        msg = 'O método split deve retornar uma lista de objetos TimeSerie'
        for ts in result:
            assert isinstance(ts, TimeSerie), msg

        return result

    @staticmethod
    def join(list_of_ts: list['TimeSerie']) -> 'TimeSerie':
        if len(list_of_ts) == 0:
            raise Exception('A lista de séries temporais não pode estar vazia')

        multivariate_ts = list_of_ts[0]
        multivariate_df = multivariate_ts.df
        timestamp_column_name = multivariate_df.columns[0]

        if len(list_of_ts) == 1:
            # Garantindo que a primeira coluna seja um Timestamp (datetime) quando o formato for long ou wide
            assert isinstance(list_of_ts[0].df[timestamp_column_name], datetime), 'timestamp deve ser do tipo datetime'
            assert list_of_ts[0].format == 'wide', 'A série temporal deve estar no formato wide'
            assert list_of_ts[0].features == 2, 'A série temporal univariada deve ter apenas 2 features'
            assert list_of_ts[0].is_univariate(), 'A série temporal deve ser univariada'
            # Garantindo que a primeira coluna seja o indice no Dataframe quando o formato for long ou wide
            if (list_of_ts[0].df).index.name == 'None':
                (list_of_ts[0].df).set_index(timestamp_column_name, inplace=True)
            return list_of_ts[0]

        # Junta os dataframes em uma série temporal multivariada
        if not isinstance(multivariate_df[timestamp_column_name][0], datetime):
            # Se a coluna timestamp não for do tipo datetime, converter para datetime.
            # Alternativamente podemeos usar o método astype(datetime) que faz o cast.
            multivariate_df[timestamp_column_name] = pd.to_datetime(multivariate_df[timestamp_column_name])

        for idx, ts in enumerate(list_of_ts):
            if idx == 0:
                continue
            if not isinstance(ts.df['timestamp'][0], datetime):
                # Se a coluna timestamp não for do tipo datetime, converter para datetime
                ts.df['timestamp'] = pd.to_datetime(ts.df['timestamp'])
                # Faz o merge dos dois Datasets, sobre a coluna timestamp
            multivariate_df = pd.merge(multivariate_df, ts.df, on='timestamp')

        # Ao final crio a série temporal multivariada usando a lista de univariadas
        features_qty = multivariate_df.columns.size
        multivariate_ts = TimeSerie(format='wide', features_qty=features_qty, data=multivariate_df)

        # Garantindo que a primeira coluna seja o indice no Dataframe quando o formato for long ou wide
        # Define a primeira coluna como índice do dataframe, caso já não seja
        if (multivariate_ts.df).index.name == 'None':
            (multivariate_ts.df).set_index(timestamp_column_name, inplace=True)
        multivariate_ts.df.set_index(timestamp_column_name, inplace=True)

        # Imprime a série temporal multivariada
        logger.debug(multivariate_ts)
        return multivariate_ts

    ### ----------------------------- Métodos de IProvenancable ----------------------------------

    def add_provenance(self, transformation: str, parameters: dict):
        # Este método adiciona informações de proveniência ao objeto TimeSerie para
        # uma transformação específica. Ele recebe como parâmetros o nome da transformação
        # e um dicionário com os parâmetros usados na transformação.
        # TODO: Implementar add_provenance()
        pass

    def get_provenances(self) -> list[dict[str, Any]]:
        # Este método retorna uma lista de dicionários com as informações de
        # proveniência de todas as transformações realizadas no objeto TimeSerie.
        # TODO: Implementar get_provenances()
        return []

    def get_provenance_by_transformation(self, transformation: str) -> dict[str, Any]:
        # Este método retorna as informações de proveniência de uma dada transformação
        # realizadas no objeto TimeSerie.
        # TODO: Implementar get_provenance_by_transformation()
        return {}

    def get_last_transformation(self, transformation: str) -> dict[str, Any]:
        # TODO: Implementar get_provenance_by_transformation()
        return {}

    def apply_transformation(self, transformation: str, **kwargs) -> Optional['TimeSerie']:
        """Este método aplica uma transformação ao objeto TimeSerie e retorna um novo
        objeto TimeSerie com as informações de proveniência atualizadas. Ele recebe como
        parâmetros o nome da transformação e os parâmetros específicos da transformação."""
        # TODO: Implementar apply_transformation()
        return TimeSerie.empty()

    ### ------------------------- Outros Métodos Estáticos da classe -----------------------------

    @staticmethod
    def empty() -> Any:
        return TimeSerie(data='EMPTY', format='long', features_qty=0)



# -*- coding: utf-8 -*-

from t8s.log_config import LogConfig
from pathlib import Path
from datetime import datetime
import numpy as np
import pandas as pd
import t8s
from t8s import get_sample_df
from t8s.ts import TimeSerie
from t8s.ts_builder import TSBuilder, ReadParquetFile, ReadCsvFile
from t8s.log_config import LogConfig
from t8s.ts_writer import TSWriter, WriteParquetFile, WriteCsvFile
from logging import INFO, DEBUG, WARNING, ERROR, CRITICAL

if __name__ == "__main__":
    # print('globals:', globals())
    # The client code.
    LogConfig().initialize_logger(DEBUG, log_file = 'logs/smoke.log')
    logger = LogConfig().getLogger()
    number_of_records = 4
    time_interval = 1 # hour
    df, last_ts = get_sample_df(number_of_records, datetime(2022, 1, 1, 0, 0, 0), time_interval)
    ts = TSBuilder.empty()
    # initialize_logger(INFO)
    logger.info('t8s package version:' + t8s.__version__)
    path_str: str = 'ts_01.parquet'
    path = Path(path_str)

    # Cria uma série temporal multivariada com três atributos: timestamp, temperatura e velocidade
    number_of_records = 4
    time_interval = 1 # hour
    dataframe, last_ts = get_sample_df(number_of_records, datetime(2022, 1, 1, 0, 0, 0), time_interval)
    ts = TimeSerie(dataframe, format='wide', features_qty=3)
    cols_str = [name for name in sorted(ts.df.columns)]
    cols_str = ', '.join(cols_str)
    logger.info(f'Dataframe com {len(ts.df.columns)} colunas: {cols_str}')
    # Faz o display da série temporal no terminal
    print(ts)

    # --------------------------------------------------------------------------------

    # Grava a série temporal em parquet
    print(f'Grava a série temporal (formato {ts.format}) em um arquivo parquet {path}')
    context = TSWriter(WriteParquetFile())
    print("Client: Strategy was seted to write Parquet file.")
    context.write(Path(path_str), ts)
    print(f'\nArquivo {str(path)} gerado à partir da TimeSerie fornecida')

    # --------------------------------------------------------------------------------

    # Lê a série temporal gravada no arquivo parquet e gera uma nova série temporal
    # com os respectivos Metadados e Schema. Ao final verifica se os tipos de dados
    # foram lidos corretamente com o módulo assert.

    # Limpando objeto ts para garantir que será lido corretamente.
    ts: TimeSerie = TSBuilder.empty()
    # Lê a série temporal gravada no arquivo parquet
    print(f'\nLendo path {str(path)} e gerando TimeSerie')

    # The client code picks a concrete strategy and passes it to the context.
    # The client should be aware of the differences between strategies in order
    # to make the right choice.
    assert isinstance(path, Path), "path must be a Path object"
    if (str(path)).endswith('.parquet'):
        context = TSBuilder(ReadParquetFile())
        print("Client: ReadStrategy is set to read Parquet file.")
        ts = context.build_from_file(Path(path_str))
    else:
        assert str(path).endswith('.csv'), "If path is not a Parquet file the path must be a CSV file"
        print("Client: ReadStrategy is set to read CSV file.")
        context = TSBuilder(ReadCsvFile())
        ts = context.build_from_file(Path(path_str))

    assert int(ts.features) == 3
    assert ts.format == 'wide'
    assert len(ts.df) == 4
    print('Tipos das colunas no Dataframe da Série Temporal:')
    for col in ts.df.columns:
        print(col, '\t', type(ts.df[col][0]))

    # pd._libs.tslibs.timestamps.Timestamp é privado e devo usar pd.Timestamp
    assert type(ts.df['timestamp'][0]) == pd.Timestamp
    assert type(ts.df['temperatura'][0]) == np.float32
    assert type(ts.df['velocidade'][0]) == np.int32

    print('---------------------------------------------------')
    univariate_list = ts.split()
    for idx, ts_uni in enumerate(univariate_list):
        print(f'TimeSerie univariada {idx+1}:')
        print(ts_uni)
        print('---------------------------------------------------')

    print('\nAgora posso fazer o join das séries temporais univariadas')
    ts = TimeSerie.join(univariate_list)
    print("\n\nTimeSerie multivariada, ts:\n")
    print(ts)

    # --------------------------------------------------------------------------------
